/*********************************************************************/
/* makeys.c (v. 0.1)                                                 */
/* Asymmetric login key generator for neblo			     */
/* Copyright (C) 2003 - Paolo Gianrossi <paolino@member.fsf.org>     */
/* 								     */
/* This program is free software; you can redistribute it and/or     */
/* modify it under the terms of the GNU General Public License as    */
/* published by the Free Software Foundation; either version 2 of    */
/* the License, or (at your option) any later version.		     */
/*								     */
/* This program is distributed in the hope that it will be useful,   */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of    */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/* General Public License for more details.			     */
/*								     */
/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software	     */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA	     */
/* 02111-1307, USA.						     */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <neblo_secret.h>

#include <time.h>


int main(int argc, char **argv)
{
  char *pubfile=DEFAULTPUBKEYFILE;
  char *secfile=DEFAULTSECKEYFILE;
  
  register int retry=3;
  
  FILE *pf, *sf;
  
  RSA *private=RSA_new();
  RSA *public=RSA_new();
  
  int bits=DEFAULTBITKEYSIZE;
  int exp=DEFAULTKEYEXP;
  int rc;
  
  int found_o=0;
  int found_b=0;
  int found_e=0;
  
  /*cmdline opts */

  for (;;) {
    while ((rc = getopt(argc,argv,"o:b:e:")) != 'o' && \
	   rc != 'b' &&rc!='e' && rc != -1);
    if (rc == -1)
      break;
    if (rc == 'b' && !found_b) /* "b" */ {
      found_b = 1;
      bits=atoi(optarg);
    }
    if (rc == 'e' && !found_e) /* "e" */ {
      found_e = 1;
      exp=atoi(optarg);
    }
    if (rc =='o' && !found_o) /* "o" */ {
      found_o = 1;
      pubfile=(char*)malloc((strlen(optarg)+5)*sizeof(char));
      secfile=(char*)malloc((strlen(optarg)+5)*sizeof(char));
      strncpy(pubfile,optarg,strlen(optarg));
      strcat(pubfile,".pk\0");
      strncpy(secfile,optarg,strlen(optarg));
      strcat(secfile,".sk\0");
    }
  }
  
  srand(time(NULL));
  while (retry > 0)
    {
      fprintf(stdout, "Generating key pair"
	      " of %d bits. Exp is %d... ", bits, exp);
      fflush(stdout);
      
      private=RSA_generate_key(bits, exp, NULL, NULL);
      
      fprintf(stdout, "done.\nChecking key validity... ");
      fflush(stdout);
      
      if(!RSA_check_key(private)){
	if(retry >1){
	  fprintf(stdout, "FAILED! Not saving keys! Retrying...\n");
	  fflush(stdout);
	  
	  retry--;
	  continue;
	} 
	else {
	  fprintf(stdout, "FAILED! Not saving keys! Not lucky..."
		  " Try again later.\n");
	  exit(1);
	}
      }
      else{
	retry = 0;
      }
    }
  
  fprintf(stdout, "done.\nSaving private key to %s... ", secfile);
  fflush(stdout);
  
  sf=fopen(secfile, "w");
  if(!sf){
    fprintf(stdout, "FAILED! Couldn't open %s! Aborting!\n", secfile);
    exit(1);
  }

/*   BIGNUM *n;              // public modulus */
/*   BIGNUM *e;              // public exponent */
/*   BIGNUM *d;              // private exponent */
/*   BIGNUM *p;              // secret prime factor */
/*   BIGNUM *q;              // secret prime factor */
/*   BIGNUM *dmp1;           // d mod (p-1) */
/*   BIGNUM *dmq1;           // d mod (q-1) */
/*   BIGNUM *iqmp;           // q^-1 mod p */

  fprintf(sf, "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s", 
	  BN_bn2hex(private->n),
	  BN_bn2hex(private->e),
	  BN_bn2hex(private->d),
	  BN_bn2hex(private->p),
	  BN_bn2hex(private->q),
	  BN_bn2hex(private->dmp1),
	  BN_bn2hex(private->dmq1),
	  BN_bn2hex(private->iqmp));
  
  fclose(sf);
  fprintf(stdout, "done\nSaving public key to %s... ", pubfile);
  fflush(stdout);
  
  public->n=private->n;
  public->e=private->e;
  pf=fopen(pubfile, "w");
  if(!pf){
    fprintf(stdout, "FAILED! Couldn't open %s! Aborting!\n", pubfile);
    exit(1);
  }
    
  fprintf(pf, "%s\n%s\n",BN_bn2hex(public->n),BN_bn2hex(public->e));
  fclose(pf);
  fprintf(stdout, "done\n--Public Key Begin--\n");
  RSA_print_fp(stdout, public, 0);
  fprintf(stdout, "--Public Key End--\n");
  return(0);
  
}
